import random
import plotly
import pandas as pd
from typing import List, Dict, Tuple, Any
import dash
import dash_table
import dash_core_components as dcc
import dash_html_components as html
from dash.dependencies import Input, Output
import plotly.graph_objs as go
import plotly.express as px

from data import load_cache, count_words_in_title, get_word_counts, get_reading_time, get_average_readed_word
from data import get_added_time_series, get_archived_time_series
from data import get_language_counts, get_favorite_count, get_domain_counts
from constants import DEFAULT_READING_SPEED


def plot_two_columns(col0: Any, col1: Any) -> html.Div:
    return html.Div(className='row', children=[
        html.Div([col0], className='two-columns--0'),
        html.Div([col1], className='two-columns--1'),
    ])


def word_cloud_plot(data: List[Dict]) -> html.Div:
    word_cnts = count_words_in_title(data)
    n_word = len(word_cnts)
    words = list(word_cnts.keys())
    weights = [word_cnts[w] for w in words]
    colors = [plotly.colors.DEFAULT_PLOTLY_COLORS[random.randrange(1, 10)] for i in range(n_word)]
    data = go.Scatter(x=[random.random() for i in range(n_word)],
                      y=[random.random() for i in range(n_word)],
                      mode='text',
                      text=words,
                      marker={'opacity': 0.3},
                      textfont={'size': weights,
                                'color': colors})
    layout = go.Layout({'xaxis': {'showgrid': False, 'showticklabels': False, 'zeroline': False},
                        'yaxis': {'showgrid': False, 'showticklabels': False, 'zeroline': False}})
    fig = go.Figure(data=[data], layout=layout)
    return html.Div([
        dcc.Graph(id='word-cloud', figure=fig)
    ])


def articles_over_time_plot(data: List[Dict], should_cumsum: bool = True) -> dcc.Graph:
    df = get_added_time_series(data)
    archived_df = get_archived_time_series(data)
    if len(archived_df) > 0:
        df = pd.merge(df, archived_df, how='outer', left_index=True, right_index=True)
    if should_cumsum:
        df.fillna(0, inplace=True)
        df = df.cumsum()
    fig = px.line(df,
                  labels={'index': 'Date', 'value': 'Number of articles'},
                  title='Article Count Over Time')
    return dcc.Graph(id='time_series', figure=fig)


def word_counts_plot(data: List[Dict]) -> dcc.Graph:
    n_last_day_options = [360, 90, 30, 7, 2]
    avg_readed_words = [int(get_average_readed_word(data, n_last_day)) for n_last_day in n_last_day_options]
    avg_readed_words_table = dash_table.DataTable(
        id='readed-words',
        columns=[{'name': f'{i} days', 'id': f'{i}_days'} for i in n_last_day_options],
        data=[{f'{i}_days': avg_readed_words[pos] for pos, i in enumerate(n_last_day_options)}],
    )
    # histogram
    fig = go.Figure()
    fig.add_trace(go.Histogram(
        x=get_word_counts(data, filters=[['status', '=', 0]]),  # unread
        name='Unread articles',
    ))
    fig.add_trace(go.Histogram(
        x=get_word_counts(data, filters=[['status', '=', 1]]),  # archived
        name='Archived articles',
    ))
    fig.update_layout(
        title_text='Word Count Distribution',  # title of plot
        xaxis_title_text='Number of words',
        yaxis_title_text='Number of articles',
        barmode='stack'  # The two histograms are drawn on top of another
    )
    return html.Div([
        html.H3(children='Average readed words recently (words / day)', className='center-text'),
        avg_readed_words_table,
        dcc.Graph(id='word-count', figure=fig)
    ])


# -------------------- Reading time -------------------- #
def get_reading_time_chart(data: List[Dict], reading_speed: int) -> go.Figure:
    fig = go.Figure()
    fig.add_trace(go.Histogram(
        x=get_reading_time(data, reading_speed=reading_speed, filters=[['status', '=', 0]]),  # unread
        name='Unread articles',
    ))
    fig.add_trace(go.Histogram(
        x=get_reading_time(data, reading_speed=reading_speed, filters=[['status', '=', 1]]),  # archived
        name='Archived articles',
    ))
    fig.update_layout(
        title_text='Reading Time Distribution',  # title of plot
        xaxis_title_text=f'Estimated reading time (minutes) with reading speed = {reading_speed} wpm',
        yaxis_title_text='Number of articles',
        barmode='stack'  # The two histograms are drawn on top of another
    )
    return fig


def get_reading_time_needed(data: List[Dict], reading_speed: int, reading_minutes_daily: int) -> html.Div:
    total_minutes = int(sum(get_reading_time(data, reading_speed=reading_speed, filters=[['status', '=', 0]])))
    days = int(total_minutes / reading_minutes_daily)
    hours = int((total_minutes % reading_minutes_daily) / 60)
    minutes = total_minutes % 60
    ans = ''
    if days > 0:
        ans += f' {days} days'
    if hours > 0:
        ans += f' {hours} hours'
    if minutes > 0:
        ans += f' {minutes} minutes'
    title = f'Total reading time needed (with {reading_speed} words / minute and {reading_minutes_daily} minutes / day)'
    return html.Div([
        html.H3(children=title, className='center-text'),
        html.Div(children=ans, className='center-text highlight'),
    ])


def reading_time_plot(data: List[Dict]) -> html.Div:
    max_reading_speed = DEFAULT_READING_SPEED * 3
    max_reading_minutes_daily = 24 * 60
    return html.Div([
        html.H3(children='Reading speed (words / minute)', className='center-text'),
        dcc.Slider(
            id='reading-speed',
            marks={i: str(i) for i in range(100, max_reading_speed, 100)},
            min=1, max=max_reading_speed, step=10, value=DEFAULT_READING_SPEED
        ),
        html.H3(children='Reading time spent daily (minutes / day)', className='center-text'),
        dcc.Slider(
            id='reading-minutes-daily',
            marks={i: str(i) for i in range(120, max_reading_minutes_daily, 240)},
            min=10, max=max_reading_minutes_daily, step=10, value=60,
        ),
        html.Div(id='reading-time-needed', children=''),
        dcc.Graph(id='reading-time', figure=go.Figure()),  # figure will be updated by update_reading_time_components()
    ])


# -------------------- Domain -------------------- #
def domain_counts_plot(data: List[Dict], limit: int = 20) -> dcc.Graph:
    top_pairs = list(get_domain_counts(data).items())  # both unread + archived
    top_pairs.sort(key=lambda p: -p[1])  # sort desc by count
    top_pairs = top_pairs[:limit]  # display top items only
    top_pairs.reverse()  # because px.bar display the items in a reversed order
    top_domains = [p[0] for p in top_pairs]
    unread_domain_cnts = get_domain_counts(data, filters=[['status', '=', 0]])
    archived_domain_cnts = get_domain_counts(data, filters=[['status', '=', 1]])
    fig = go.Figure()
    fig.add_trace(go.Bar(
        x=[unread_domain_cnts.get(d, 0) for d in top_domains],
        y=top_domains,
        name='Unread articles',
        orientation='h',
    ))
    fig.add_trace(go.Bar(
        x=[archived_domain_cnts.get(d, 0) for d in top_domains],
        y=top_domains,
        name='Archived articles',
        orientation='h',
    ))
    fig.update_layout(
        title_text='Top Domains',
        barmode='stack',
        yaxis=dict(tickmode='linear'),  # to show ALL labels
    )
    return dcc.Graph(id='domain-counts', figure=fig)


def language_counts_plot(data: List[Dict]) -> dcc.Graph:
    pairs = list(get_language_counts(data).items())
    fig = go.Figure(
        data=[
            go.Pie(
                labels=[p[0] for p in pairs],
                values=[p[1] for p in pairs],
            ),
        ],
        layout_title_text="Languages",
    )
    return dcc.Graph(id='language-counts', figure=fig)


def favorite_count_plot(data: List[Dict]) -> html.Div:
    res = get_favorite_count(data)
    return html.Div(
        [
            html.H2(children='Favorite', className='center-text'),
            html.H2(
                children=f"{res['count']} articles ({'%.2f' % (100.0 * res['percent'])} %)",
                className='center-text highlight',
            )
        ],
        className='favorite-div',
    )


def create_app(data: List[Dict] = None) -> dash.Dash:
    if data is None:
        data = load_cache()
    app = dash.Dash()
    app.title = "Pocket Stats"
    app.layout = html.Div(style={}, children=[
        word_cloud_plot(data),
        articles_over_time_plot(data),
        plot_two_columns(word_counts_plot(data), reading_time_plot(data)),
        domain_counts_plot(data),
        plot_two_columns(language_counts_plot(data), favorite_count_plot(data)),
    ])

    @app.callback(
        [Output(component_id='reading-time', component_property='figure'),
         Output(component_id='reading-time-needed', component_property='children')],
        [Input(component_id='reading-speed', component_property='value'),
         Input(component_id='reading-minutes-daily', component_property='value')]
    )
    def update_reading_time_components(reading_speed: int, reading_minutes_daily: int) -> Tuple[go.Figure, str]:
        return (get_reading_time_chart(data, reading_speed),
                get_reading_time_needed(data, reading_speed, reading_minutes_daily))

    return app



import click
from data import fetch_data as _fetch_data
from visualization import create_app


@click.command()
@click.option('--offset', type=int, default=0, help='First item position to be fetched')
@click.option('--limit', type=int, default=None, help='Number of items to be fetched')
@click.option('--overwrite_cache', is_flag=True, help='Will overwrite the local cache')
def fetch_data(offset: int, limit: int, overwrite_cache: bool) -> None:
    ans = _fetch_data(offset, limit, overwrite_cache)
    if len(ans) > 0:
        print('Sample record:')
        print(ans[0])


@click.command()
@click.option('--debug', is_flag=True, help='Debug mode')
@click.option('--port', type=int, default=8050, help='Port of the web server. Default = 8050.')
def webapp(debug: bool, port: int) -> None:
    app = create_app()
    app.run_server(debug=debug, port=port)


@click.group()
def cli() -> None:
    pass


cli.add_command(fetch_data)
cli.add_command(webapp)


if __name__ == '__main__':
    cli()



import sys
import logging
import os
import pytz


DEFAULT_TZINFO = pytz.utc
CONSUMER_KEY = os.environ.get('POCKET_STATS_CONSUMER_KEY', None)
ACCESS_TOKEN = os.environ.get('POCKET_STATS_ACCESS_TOKEN', None)
CACHE_FILE = os.environ.get('POCKET_STATS_CACHE_FILE', os.path.expanduser('~/pocket-tools.cache'))
DEFAULT_READING_SPEED = 225  # words per minute


root = logging.getLogger()
root.setLevel(logging.INFO)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - [%(levelname)s] - %(name)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)



import nltk
import errno
import os
import json
import logging
import tldextract
from datetime import datetime, timedelta
from pocket import Pocket
from typing import List, Dict
from collections import Counter
from nltk.corpus import stopwords
import pandas as pd
from constants import CACHE_FILE, CONSUMER_KEY, ACCESS_TOKEN, DEFAULT_READING_SPEED, DEFAULT_TZINFO


# ------- Helper functions ------- #
def download_ntlk():
    try:
        nltk.data.find('corpora/stopwords')
    except LookupError:
        nltk.download('stopwords')


download_ntlk()
invalid_words = stopwords.words('english')


def is_valid_word(w):
    if len(w) == 0:
        return False
    return w not in invalid_words


def normalize_language_name(lang: str) -> str:
    lang = lang.strip()
    if len(lang) == 0:
        return 'unknown'
    return lang


def get_domain_from_url(url: str) -> str:
    extract_result = tldextract.extract(url)
    return extract_result.domain + '.' + extract_result.suffix


def epoch_to_yyymmdd(epoch: int):
    return datetime.fromtimestamp(int(epoch), tz=DEFAULT_TZINFO).strftime('%Y%m%d')


# ------- Main functions ------- #
def fetch_data(offset: int = 0, limit: int = None, overwrite_cache: bool = False,
               consumer_key: str = CONSUMER_KEY, access_token: str = ACCESS_TOKEN) -> List[Dict]:
    assert (consumer_key is not None) and (access_token is not None), 'Please set value for these environment variables'
    api = Pocket(consumer_key=consumer_key, access_token=access_token)
    ans = []
    count = 500
    if limit is not None:
        limit += offset
        count = min(count, limit)
    while (limit is None) or (offset < limit):
        response = api.retrieve(
            offset=offset, count=count,
            state='all',  # both unread and archived items
        )
        items = response.get('list', {})
        n = len(items)
        if n == 0:
            break
        ans.extend(item for k, item in items.items())
        logging.info(f'Fetched {n} records. Total = {len(ans)} now.')
        offset += n
    if overwrite_cache:
        logging.info(f'Writing data to cache file {CACHE_FILE}')
        with open(CACHE_FILE, 'w') as fo:
            json.dump(ans, fo)
    return ans


def load_cache(cache_file: str = CACHE_FILE) -> List[Dict]:
    if not os.path.isfile(cache_file):
        logging.error("Missing cache file, please run 'python -m pocket_stats fetch-data --overwrite_cache'")
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                cache_file)
    with open(cache_file, 'r') as fi:
        return json.load(fi)


# filter format: [key, operation, expected_value]
def should_pass_filter(filter: List, record: Dict):
    key, op, expected = filter
    expected = str(expected) if expected is not None else None
    value = record.get(key, None)
    value = str(value) if value is not None else None
    if op == '=':
        return value == expected
    elif op == '!=':
        return value != expected
    else:
        raise NotImplementedError


def should_pass_filters(filters: List[List], record: Dict) -> bool:
    for f in filters:
        if not should_pass_filter(f, record):
            return False
    return True


def count_words_in_title(data: List[Dict]) -> Dict[str, int]:
    words = []
    for record in data:
        title = record['given_title']
        words.extend(x.strip().lower() for x in title.split(' ') if is_valid_word(x.strip().lower()))
    return Counter(words)


def get_word_counts(data: List[Dict], filters: List[List] = []) -> List[int]:
    return [int(record.get('word_count', -99999))
            for record in data
            if should_pass_filters(filters, record)]


def get_reading_time(data: List[Dict],
                     reading_speed: int = DEFAULT_READING_SPEED,
                     filters: List[List] = []) -> List[float]:
    # because some records in data don't have the 'time_to_read' field
    word_counts = get_word_counts(data, filters=filters)
    return [wc / reading_speed for wc in word_counts if wc > 0]


def get_added_time_series(data: List[Dict]) -> pd.DataFrame:
    added_date_counts = Counter(epoch_to_yyymmdd(record['time_added']) for record in data)
    df = pd.DataFrame.from_dict({datetime.strptime(d, '%Y%m%d'): cnt for d, cnt in added_date_counts.items()},
                                orient='index', columns=['All articles'])
    df.index = df.index.tz_localize('UTC')
    return df


def get_archived_time_series(data: List[Dict]) -> pd.DataFrame:
    archived_date_counts = Counter(
        epoch_to_yyymmdd(record['time_added']) for record in data
        if int(record['status']) == 1
    )
    df = pd.DataFrame.from_dict({datetime.strptime(d, '%Y%m%d'): cnt for d, cnt in archived_date_counts.items()},
                                orient='index', columns=['Archived articles'])
    df.index = df.index.tz_localize('UTC')
    return df


def get_average_readed_word(data: List[Dict], n_last_days: int) -> float:
    # find the ones that are archived, time_updated >= min_date
    min_date = datetime.now() - timedelta(days=n_last_days)
    records = [
        record for record in data
        if (int(record['status']) == 1) and (datetime.fromtimestamp(int(record['time_updated'])) >= min_date)
    ]
    word_counts = get_word_counts(records)
    return 0 if len(word_counts) == 0 else sum(word_counts) / len(word_counts)


def get_domain_counts(data: List[Dict], filters: List[List] = []) -> Dict[str, int]:
    return Counter(get_domain_from_url(record['resolved_url'])
                   for record in data
                   if should_pass_filters(filters, record))


def get_language_counts(data: List[Dict]) -> Dict[str, int]:
    return Counter(normalize_language_name(record['lang']) for record in data)


def get_favorite_count(data: List[Dict]) -> Dict[str, int]:
    total = len(data)
    cnt = sum(1 for record in data if int(record['favorite']) == 1)
    return {
        'count': cnt,
        'percent': 1.0 * cnt / total if total > 0 else 0,
    }


def get_unread_count(data: List[Dict]) -> int:
    df = pd.DataFrame.from_dict(data)
    return df.query("status == '0'").status.count()



'''
  setup.py
'''

from os import path
from setuptools import setup


# read the contents of your README file
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='pocket_stats',
    version='0.2.2',
    description='Tools for the Pocket reading app https://app.getpocket.com/',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/nlbao/pocket_stats.git',
    author='Bao Nguyen',
    author_email='nlbao95@gmail.com',
    license='MIT',
    packages=['pocket_stats'],
    install_requires=[
        'setuptools',
        'click',
        'pocket-api',
        'dash',
        'dash-renderer',
        'dash-html-components',
        'dash-core-components',
        'plotly',
        'nltk',
        'tldextract',
        'pandas',
    ],
    tests_require=['pytest', 'pytest-cov', 'freezegun'],
    zip_safe=False
)
